# from typing import Optional, Any, List, Dict
#
# from botocore.exceptions import ClientError
#
# from phidata.infra.aws.api_client import AwsApiClient
# from phidata.infra.aws.resource.base import AwsResource
# from phidata.infra.aws.resource.iam.policy import IamPolicy
# from phidata.utils.cli_console import print_info, print_error
# from phidata.utils.log import logger
#
#
# class IamGroup(AwsResource):
#     """
#     # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/iam.html#group
#     """
#
#     resource_type = "IamGroup"
#     service_name = "iam"
#
#     # GroupName: The name of the group to create.
#     name: str
#     # The trust relationship policy document that grants an entity permission to assume the group.
#     assume_role_policy_document: str
#     # The path to the group. This parameter is optional. If it is not included, it defaults to a slash (/).
#     path: Optional[str] = None
#     # A description of the group.
#     description: Optional[str] = None
#     # The maximum session duration (in seconds) that you want to set for the specified group.
#     # If you do not specify a value for this setting, the default maximum of one hour is applied.
#     # This setting can have a value from 1 hour to 12 hours.
#     max_session_duration: Optional[int] = None
#     # The ARN of the policy that is used to set the permissions boundary for the group.
#     permissions_boundary: Optional[str] = None
#     # A list of tags that you want to attach to the new group. Each tag consists of a key name and an associated value.
#     tags: Optional[List[Dict[str, str]]] = None
#
#     # List of IAM policies to
#     # attach to the group after it is created
#     policies: Optional[List[IamPolicy]] = None
#     # List of IAM policy ARNs (Amazon Resource Name) to
#     # attach to the group after it is created
#     policy_arns: Optional[List[str]] = None
#
#     # The Amazon Resource Name (ARN) specifying the group.
#     # To get the arn, use get_arn() function
#     arn: Optional[str] = None
#
#     def _create(self, aws_client: AwsApiClient) -> bool:
#         """Creates the IamGroup
#
#         Args:
#             aws_client: The AwsApiClient for the current cluster
#         """
#
#         print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
#         try:
#             # create a dict of args which are not null, otherwise aws type validation fails
#             not_null_args: Dict[str, Any] = {}
#             if self.path:
#                 not_null_args["Path"] = self.path
#             if self.description:
#                 not_null_args["Description"] = self.description
#             if self.max_session_duration:
#                 not_null_args["MaxSessionDuration"] = self.max_session_duration
#             if self.permissions_boundary:
#                 not_null_args["PermissionsBoundary"] = self.permissions_boundary
#             if self.tags:
#                 not_null_args["Tags"] = self.tags
#
#             # Create Group
#             service_resource = self.get_service_resource(aws_client)
#             group = service_resource.create_role(
#                 GroupName=self.name,
#                 AssumeGroupPolicyDocument=self.assume_role_policy_document,
#                 **not_null_args,
#             )
#             # logger.debug(f"Group: {group}")
#
#             # Validate Group creation
#             create_date = group.create_date
#             self.arn = group.arn
#             logger.debug(f"create_date: {create_date}")
#             logger.debug(f"arn: {self.arn}")
#             if create_date is not None:
#                 print_info(f"Group created: {self.name}")
#                 self.active_resource = group
#                 return True
#             print_error("Group could not be created")
#         except Exception as e:
#             print_error(f"{self.get_resource_type()} could not be created.")
#             print_error(e)
#         return False
#
#     def post_create(self, aws_client: AwsApiClient) -> bool:
#         # Wait for Group to be created
#         if self.wait_for_creation:
#             try:
#                 print_info(f"Waiting for {self.get_resource_type()} to be created.")
#                 waiter = self.get_service_client(aws_client).get_waiter("role_exists")
#                 waiter.wait(
#                     GroupName=self.name,
#                     WaiterConfig={
#                         "Delay": self.waiter_delay,
#                         "MaxAttempts": self.waiter_max_attempts,
#                     },
#                 )
#             except Exception as e:
#                 print_error("Waiter failed.")
#                 print_error(e)
#         # Attach policy arns to group
#         if self.active_resource is not None and self.policy_arns is not None:
#             self.attach_policy_arns(aws_client)
#         # Attach policies to group
#         if self.active_resource is not None and self.policies is not None:
#             self.attach_policies(aws_client)
#         return True
#
#     def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
#         """Returns the IamGroup
#
#         Args:
#             aws_client: The AwsApiClient for the current cluster
#         """
#         logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
#         try:
#             service_resource = self.get_service_resource(aws_client)
#             group = service_resource.Group(name=self.name)
#             group.load()
#             create_date = group.create_date
#             self.arn = group.arn
#             logger.debug(f"create_date: {create_date}")
#             logger.debug(f"arn: {self.arn}")
#             if create_date is not None:
#                 logger.debug(f"Group found: {group.role_name}")
#                 self.active_resource = group
#         except ClientError as ce:
#             logger.debug(f"ClientError: {ce}")
#             pass
#         except Exception as e:
#             print_error(f"Error reading {self.get_resource_type()}.")
#             print_error(e)
#         return self.active_resource
#
#     def _delete(self, aws_client: AwsApiClient) -> bool:
#         """Deletes the IamGroup
#
#         Args:
#             aws_client: The AwsApiClient for the current cluster
#         """
#
#         print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
#         try:
#             group = self._read(aws_client)
#             # logger.debug(f"Group: {group}")
#             # logger.debug(f"Group type: {type(group)}")
#             self.active_resource = None
#
#             if group is None:
#                 logger.warning(f"No {self.get_resource_type()} to delete")
#                 return True
#
#             # detach all policies
#             policies = group.attached_policies.all()
#             for policy in policies:
#                 print_info(f"Detaching policy: {policy}")
#                 group.detach_policy(PolicyArn=policy.arn)
#
#             # detach all instance profiles
#             profiles = group.instance_profiles.all()
#             for profile in profiles:
#                 print_info(f"Removing group from profile: {profile}")
#                 profile.remove_role(GroupName=group.name)
#
#             # delete group
#             group.delete()
#             print_info(
#                 f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
#             )
#             return True
#         except Exception as e:
#             print_error(f"{self.get_resource_type()} could not be deleted.")
#             print_error("Please try again or delete resources manually.")
#             print_error(e)
#         return False
#
#     def attach_policy_arns(self, aws_client: AwsApiClient) -> bool:
#         """
#         Attaches the specified managed policy to the specified IAM group.
#         When you attach a managed policy to a group, the managed policy becomes part of the
#         group's permission (access) policy.
#
#         Returns:
#             True if operation was successful
#         """
#         if self.policy_arns is None:
#             return True
#
#         group = self._read(aws_client)
#         if group is None:
#             logger.warning(f"No {self.get_resource_type()} to attach")
#             return True
#         try:
#             # logger.debug("Attaching managed policies to group")
#             for arn in self.policy_arns:
#                 if isinstance(arn, str):
#                     group.attach_policy(PolicyArn=arn)
#                     print_info(f"Attaching policy to {group.role_name}: {arn}")
#             return True
#         except Exception as e:
#             print_error(e)
#         return False
#
#     def attach_policies(self, aws_client: AwsApiClient) -> bool:
#         """
#         Returns:
#             True if operation was successful
#         """
#         if self.policies is None:
#             return True
#
#         group = self._read(aws_client)
#         if group is None:
#             logger.warning(f"No {self.get_resource_type()} to attach")
#             return True
#         try:
#             logger.debug("Attaching managed policies to group")
#             for policy in self.policies:
#                 if policy.arn is None:
#                     policy.create(aws_client)
#                 if policy.arn is not None:
#                     group.attach_policy(PolicyArn=policy.arn)
#                     print_info(f"Attaching policy to {group.role_name}: {policy.arn}")
#             return True
#         except Exception as e:
#             print_error(e)
#         return False
#
#     def get_arn(self, aws_client: AwsApiClient) -> Optional[str]:
#         group = self._read(aws_client)
#         if group is None:
#             return None
#
#         self.arn = group.arn
#         return self.arn



from typing import Dict, List, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.enums.storage_class import StorageClassType
from phidata.infra.k8s.resource.storage_k8s_io.v1.storage_class import StorageClass
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.infra.k8s.exceptions import StorageClassNotFoundException
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class CreateStorageClass(BaseModel):
    storage_class_name: str
    app_name: str
    storage_class_type: Optional[StorageClassType] = None
    parameters: Optional[Dict[str, str]] = None
    provisioner: Optional[str] = None
    allow_volume_expansion: Optional[str] = None
    mount_options: Optional[List[str]] = None
    reclaim_policy: Optional[str] = None
    volume_binding_mode: Optional[str] = None
    namespace: Optional[str] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[StorageClass]:
        """Creates a StorageClass resource."""

        logger.debug(f"Init StorageClass resource: {self.storage_class_name}")
        sc_labels = create_component_labels_dict(
            component_name=self.storage_class_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        # construct the provisioner and parameters
        sc_provisioner: str
        sc_parameters: Dict[str, str]

        # if the provisioner is provided, use that
        if self.provisioner is not None:
            sc_provisioner = self.provisioner
        # otherwise derive the provisioner from the StorageClassType
        elif self.storage_class_type is not None:
            if self.storage_class_type in (
                StorageClassType.GCE_SSD,
                StorageClassType.GCE_STANDARD,
            ):
                sc_provisioner = "kubernetes.io/gce-pd"
            else:
                raise StorageClassNotFoundException(
                    f"{self.storage_class_type} not found"
                )
        else:
            print_error(
                f"No provisioner or StorageClassType found for {self.storage_class_name}"
            )
            return None

        # if the parameters are provided use those
        if self.parameters is not None:
            sc_parameters = self.parameters
        # otherwise derive the parameters from the StorageClassType
        elif self.storage_class_type is not None:
            if self.storage_class_type == StorageClassType.GCE_SSD:
                sc_parameters = {"type": "pd-ssd"}
            if self.storage_class_type == StorageClassType.GCE_STANDARD:
                sc_parameters = {"type": "pd-standard"}
            else:
                raise StorageClassNotFoundException(
                    f"{self.storage_class_type} not found"
                )
        else:
            print_error(
                f"No parameters or StorageClassType found for {self.storage_class_name}"
            )
            return None

        _storage_class = StorageClass(
            api_version=ApiVersion.STORAGE_V1,
            kind=Kind.STORAGECLASS,
            metadata=ObjectMeta(
                name=self.storage_class_name,
                labels=sc_labels,
            ),
            allow_volume_expansion=self.allow_volume_expansion,
            mount_options=self.mount_options,
            provisioner=sc_provisioner,
            parameters=sc_parameters,
            reclaim_policy=self.reclaim_policy,
            volume_binding_mode=self.volume_binding_mode,
        )

        logger.debug(
            f"StorageClass {self.storage_class_name}:\n{_storage_class.json(exclude_defaults=True, indent=2)}"
        )
        return _storage_class



from typing import Optional, Dict

from docker.models.containers import Container
from docker.errors import APIError

from phidata.utils.log import logger


def execute_command(
    cmd: str,
    container: Container,
    wait: bool = True,
    detach: bool = False,
    print_output: bool = True,
    docker_env: Optional[Dict[str, str]] = None,
) -> bool:
    """

    :param cmd:
    :param container:
    :param wait:
    :param detach:
    :param print_output:
    :param docker_env:
    :return:

    TODO:
    * Run a detached command
    * Return False on failure, currently only True is returned
    """
    import os
    import socket
    import time

    try:
        logger.debug("Exec `{}` in container: {}".format(cmd, container.name))
        _container_socket: Optional[socket.SocketIO] = None
        exit_code, _container_socket = container.exec_run(
            cmd=cmd,
            stdout=True,
            stdin=True,
            tty=True,
            socket=True,
            environment=docker_env,
        )
        # logger.debug(f"exit_code: {exit_code}")
        # logger.debug(f"_container_socket: {_container_socket}")
        if _container_socket is None:
            logger.debug("container_socket is None")
            return False

        if not (wait or print_output):
            return True

        if _container_socket.readable():
            # code references the following
            # - https://stackoverflow.com/a/66329671
            # logger.debug("_container_socket is readable")
            _container_socket._sock.setblocking(False)  # type: ignore

            _op = None
            while True:
                try:
                    # Read up-to 10000 bytes.
                    _op = os.read(_container_socket.fileno(), 10000)
                    # logger.debug("Op: {}".format(_op))
                except BlockingIOError as not_ready_error:
                    # logger.exception(not_ready_error)
                    logger.info("waiting...")
                    time.sleep(5)
                    continue

                if _op is None:
                    logger.info("Output not available.")
                    break
                if (
                    _op == b""
                    or (isinstance(_op, bytes) and _op.decode(errors="ignore")) == ""
                ):
                    logger.info("Task finished")
                    break

                # Use print here because console.print has issues decoding output
                if print_output:
                    print(_op.decode(errors="ignore"))
        return True
    except APIError as api_error:
        logger.error(f"APIError: {api_error}")
    return False


# def execute_command_stream(
#     cmd: str,
#     container: Container,
#     wait: bool = True,
#     detach: bool = False,
#     print_output: bool = True,
#     docker_env: Optional[Dict[str, str]] = None,
# ) -> bool:
#     """
#     * https://docker-py.readthedocs.io/en/stable/user_guides/multiplex.html
#     """
#     import os
#     import time
#
#     try:
#         logger.debug("Exec `{}` in container: {}".format(cmd, container.name))
#         _, stream = container.exec_run(
#             cmd=cmd,
#             stdout=True,
#             stdin=True,
#             tty=True,
#             stream=True,
#             demux=False,
#             environment=docker_env,
#         )
#         for data in stream:
#             logger.info(f"data: {data.decode()}")
#     except APIError as api_error:
#         logger.error(f"APIError: {api_error}")
#     return False



from typing import List, Optional

from pydantic import validator

from phidata.app.phidata_app import PhidataApp
from phidata.app.databox import default_databox_name
from phidata.infra.base import InfraArgs
from phidata.infra.docker.resource.image import DockerImage
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.volume import DockerVolume
from phidata.infra.docker.resource.group import DockerResourceGroup


class DockerArgs(InfraArgs):
    network: str = "phi"
    # configure the DockerClient
    # URL to the Docker server. For example, unix:///var/run/docker.sock or tcp://127.0.0.1:1234.
    # when None, phidata use DockerClient.from_env()
    base_url: Optional[str] = None
    apps: Optional[List[PhidataApp]] = None
    # default resources
    images: Optional[List[DockerImage]] = None
    containers: Optional[List[DockerContainer]] = None
    volumes: Optional[List[DockerVolume]] = None
    # additional resource groups
    resources: Optional[List[DockerResourceGroup]] = None
    databox_name: str = default_databox_name

    @validator("apps")
    def apps_are_valid(cls, apps):
        if apps is not None:
            for _app in apps:
                if not isinstance(_app, PhidataApp):
                    raise TypeError("App not of type PhidataApp: {}".format(_app))
        return apps

    def default_resources_available(self) -> bool:
        return (
            self.images is not None
            or self.containers is not None
            or self.volumes is not None
        )



from phidata.task import Task, TaskArgs
from phidata.utils.log import logger


class CheckArgs(TaskArgs):
    pass


class Check(Task):
    """Base Class for all DQChecks"""

    def __init__(self):
        super().__init__()



from typing import Any, Optional, Dict

from phidata.infra.base.resource import InfraResource
from phidata.infra.docker.api_client import DockerApiClient
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class DockerResource(InfraResource):
    """Base class for all Docker Resources."""

    # Fields specific to docker resources

    # Fields for the resource from the DockerApiClient
    id: Optional[str] = None
    short_id: Optional[str] = None
    attrs: Optional[Dict[str, Any]] = None

    @staticmethod
    def get_from_cluster(docker_client: DockerApiClient) -> Any:
        logger.error(f"@get_from_cluster method not defined")
        return None

    def _create(self, docker_client: DockerApiClient) -> bool:
        logger.error(f"@_create method not defined for {self.__class__.__name__}")
        return False

    def create(self, docker_client: DockerApiClient) -> bool:
        """Creates the resource on the docker cluster

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        if self.use_cache and self.is_active_on_cluster(docker_client):
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} already active."
            )
            return True
        return self._create(docker_client)

    def _read(self, docker_client: DockerApiClient) -> Any:
        logger.error(f"@_read method not defined for {self.__class__.__name__}")
        return False

    def read(self, docker_client: DockerApiClient) -> Any:
        """Reads the resource from the docker cluster
        Eg:
            * For a Network resource, it will return the DockerNetwork object
            currently running on docker.

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        if self.use_cache and self.active_resource is not None:
            return self.active_resource
        return self._read(docker_client)

    def _update(self, docker_client: DockerApiClient) -> bool:
        logger.error(f"@_update method not defined for {self.__class__.__name__}")
        return False

    def update(self, docker_client: DockerApiClient) -> bool:
        """Updates the resource on the docker cluster

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        if self.is_active_on_cluster(docker_client):
            return self._update(docker_client)
        else:
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} not active, creating"
            )
            return self.create(docker_client=docker_client)

    def _delete(self, docker_client: DockerApiClient) -> bool:
        logger.error(f"@_delete method not defined for {self.__class__.__name__}")
        return False

    def delete(self, docker_client: DockerApiClient) -> bool:
        """Deletes the resource from the docker cluster

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        # Skip resource deletion if skip_delete = True
        if self.skip_delete:
            print_info(f"Skipping delete: {self.get_resource_name()}")
            return True
        if self.is_active_on_cluster(docker_client):
            return self._delete(docker_client)
        else:
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} not active."
            )
            return True

    def is_active_on_cluster(self, docker_client: DockerApiClient) -> bool:
        """Returns True if the resource is running on the docker cluster"""
        active_resource = self.read(docker_client=docker_client)
        if active_resource is not None:
            return True
        return False



from typing import Dict, List, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.enums.pv import PVAccessMode
from phidata.infra.k8s.resource.core.v1.persistent_volume_claim import (
    PersistentVolumeClaim,
    PersistentVolumeClaimSpec,
)
from phidata.infra.k8s.resource.core.v1.resource_requirements import (
    ResourceRequirements,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.log import logger


class CreatePVC(BaseModel):
    pvc_name: str
    app_name: str
    namespace: Optional[str] = None
    request_storage: str
    storage_class_name: str
    access_modes: List[PVAccessMode] = [PVAccessMode.READ_WRITE_ONCE]
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[PersistentVolumeClaim]:
        """Creates a PersistentVolumeClaim resource."""

        pvc_name = self.pvc_name
        logger.debug(f"Init PersistentVolumeClaim resource: {pvc_name}")

        pvc_labels = create_component_labels_dict(
            component_name=pvc_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        pvc = PersistentVolumeClaim(
            api_version=ApiVersion.CORE_V1,
            kind=Kind.PERSISTENTVOLUMECLAIM,
            metadata=ObjectMeta(
                name=pvc_name,
                namespace=self.namespace,
                labels=pvc_labels,
            ),
            spec=PersistentVolumeClaimSpec(
                access_modes=self.access_modes,
                resources=ResourceRequirements(
                    requests={"storage": self.request_storage}
                ),
                storage_class_name=self.storage_class_name,
            ),
        )

        # logger.info(
        #     f"PersistentVolumeClaim {pvc_name}:\n{pvc.json(exclude_defaults=True, indent=2)}"
        # )
        return pvc



from typing import Optional

from phidata.asset.file import File
from phidata.asset.aws.s3.object import S3Object
from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs


class DownloadS3ObjectToFileArgs(PythonTaskArgs):
    file: File
    s3_object: S3Object


class DownloadS3ObjectToFile(PythonTask):
    def __init__(
        self,
        file: File,
        s3_object: S3Object,
        name: str = "download_s3_object_to_file",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: DownloadS3ObjectToFileArgs = DownloadS3ObjectToFileArgs(
                file=file,
                s3_object=s3_object,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=download_s3_object_to_file,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def file(self) -> File:
        return self.args.file

    @file.setter
    def file(self, file: File) -> None:
        if file is not None:
            self.args.file = file

    @property
    def s3_object(self) -> S3Object:
        return self.args.s3_object

    @s3_object.setter
    def s3_object(self, s3_object: S3Object) -> None:
        if s3_object is not None:
            self.args.s3_object = s3_object


def download_s3_object_to_file(**kwargs) -> bool:

    args: DownloadS3ObjectToFileArgs = DownloadS3ObjectToFileArgs(**kwargs)
    # logger.debug("DownloadS3ObjectToFileArgs: {}".format(args))

    if args.file is None:
        logger.error("File not available")
        return False

    if args.s3_object is None:
        logger.error("S3Object not available")
        return False

    download_success = args.s3_object.download_file(file=args.file)
    return download_success



from typing import Optional, Any, Dict, List

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource, AwsObject
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class GlueS3Target(AwsObject):
    # The directory path in the S3 bucket to target
    dir: str = ""
    # The s3 bucket to target
    bucket: S3Bucket
    # A list of glob patterns used to exclude from the crawl.
    # For more information, see https://docs.aws.amazon.com/glue/latest/dg/add-crawler.html
    exclusions: Optional[List[str]] = None
    # The name of a connection which allows a job or crawler to access data in Amazon S3 within an
    # Amazon Virtual Private Cloud environment (Amazon VPC).
    connection_name: Optional[str] = None
    # Sets the number of files in each leaf folder to be crawled when crawling sample files in a dataset.
    # If not set, all the files are crawled. A valid value is an integer between 1 and 249.
    sample_size: Optional[int] = None
    # A valid Amazon SQS ARN. For example, arn:aws:sqs:region:account:sqs .
    event_queue_arn: Optional[str] = None
    # A valid Amazon dead-letter SQS ARN. For example, arn:aws:sqs:region:account:deadLetterQueue .
    dlq_event_queue_arn: Optional[str] = None


class GlueCrawler(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glue.html
    """

    resource_type = "GlueCrawler"
    service_name = "glue"

    # Name of the crawler.
    name: str
    # The IAM role for the crawler
    iam_role: IamRole
    # List of GlueS3Target to add to the targets dict
    s3_targets: Optional[List[GlueS3Target]] = None
    # The Glue database where results are written,
    # such as: arn:aws:daylight:us-east-1::database/sometable/* .
    database_name: Optional[str] = None
    # A description of the new crawler.
    description: Optional[str] = None
    # A list of collection of targets to crawl.
    targets: Optional[Dict[str, List[dict]]] = None
    # A cron expression used to specify the schedule
    # For example, to run something every day at 12:15 UTC,
    # you would specify: cron(15 12 * * ? *) .
    schedule: Optional[str] = None
    # A list of custom classifiers that the user has registered.
    # By default, all built-in classifiers are included in a crawl,
    # but these custom classifiers always override the default classifiers for a given classification.
    classifiers: Optional[List[str]] = None
    # The table prefix used for catalog tables that are created.
    table_prefix: Optional[str] = None
    # The policy for the crawler's update and deletion behavior.
    schema_change_policy: Optional[Dict[str, str]] = None
    # A policy that specifies whether to crawl the entire dataset again,
    # or to crawl only folders that were added since the last crawler run.
    recrawl_policy: Optional[Dict[str, str]] = None
    lineage_configuration: Optional[Dict[str, str]] = None
    lake_formation_configuration: Optional[Dict[str, str]] = None
    # Crawler configuration information. This versioned JSON string
    # allows users to specify aspects of a crawler's behavior.
    configuration: Optional[str] = None
    # The name of the SecurityConfiguration structure to be used by this crawler.
    crawler_security_configuration: Optional[str] = None
    # The tags to use with this crawler request.
    tags: Optional[Dict[str, str]] = None

    # provided by api on create
    creation_time: Optional[str] = None
    last_crawl: Optional[str] = None

    def get_glue_crawler_targets(self) -> Optional[Dict[str, List[dict]]]:
        # start with user provided targets
        crawler_targets: Optional[Dict[str, List[dict]]] = self.targets

        # Add GlueS3Targets to crawler_targets
        if self.s3_targets is not None:
            # create S3Targets dicts using s3_targets
            new_s3_targets_list: List[dict] = []
            for s3_target in self.s3_targets:
                _new_s3_target_path = f"s3://{s3_target.bucket.name}/{s3_target.dir}"
                # start with the only required argument
                _new_s3_target_dict: Dict[str, Any] = {"Path": _new_s3_target_path}
                # add any optional arguments
                if s3_target.exclusions is not None:
                    _new_s3_target_dict["Exclusions"] = s3_target.exclusions
                if s3_target.connection_name is not None:
                    _new_s3_target_dict["ConnectionName"] = s3_target.connection_name
                if s3_target.sample_size is not None:
                    _new_s3_target_dict["SampleSize"] = s3_target.sample_size
                if s3_target.event_queue_arn is not None:
                    _new_s3_target_dict["EventQueueArn"] = s3_target.event_queue_arn
                if s3_target.dlq_event_queue_arn is not None:
                    _new_s3_target_dict[
                        "DlqEventQueueArn"
                    ] = s3_target.dlq_event_queue_arn

                new_s3_targets_list.append(_new_s3_target_dict)

            # Add new S3Targets to crawler_targets
            if crawler_targets is None:
                crawler_targets = {}
            # logger.debug(f"new_s3_targets_list: {new_s3_targets_list}")
            existing_s3_targets = crawler_targets.get("S3Targets", [])
            # logger.debug(f"existing_s3_targets: {existing_s3_targets}")
            new_s3_targets = existing_s3_targets + new_s3_targets_list
            # logger.debug(f"new_s3_targets: {new_s3_targets}")
            crawler_targets["S3Targets"] = new_s3_targets

        # TODO: add more targets as needed
        logger.debug(f"GlueCrawler targets: {crawler_targets}")
        return crawler_targets

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the GlueCrawler

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        from botocore.exceptions import ClientError

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}
            if self.database_name:
                not_null_args["DatabaseName"] = self.database_name
            if self.description:
                not_null_args["Description"] = self.description
            if self.schedule:
                not_null_args["Schedule"] = self.schedule
            if self.classifiers:
                not_null_args["Classifiers"] = self.classifiers
            if self.table_prefix:
                not_null_args["TablePrefix"] = self.table_prefix
            if self.schema_change_policy:
                not_null_args["SchemaChangePolicy"] = self.schema_change_policy
            if self.recrawl_policy:
                not_null_args["RecrawlPolicy"] = self.recrawl_policy
            if self.lineage_configuration:
                not_null_args["LineageConfiguration"] = self.lineage_configuration
            if self.lake_formation_configuration:
                not_null_args[
                    "LakeFormationConfiguration"
                ] = self.lake_formation_configuration
            if self.configuration:
                not_null_args["Configuration"] = self.configuration
            if self.crawler_security_configuration:
                not_null_args[
                    "CrawlerSecurityConfiguration"
                ] = self.crawler_security_configuration
            if self.tags:
                not_null_args["Tags"] = self.tags

            targets = self.get_glue_crawler_targets()
            if targets:
                not_null_args["Targets"] = targets

            # Create crawler
            # Get the service_client
            service_client = self.get_service_client(aws_client)
            iam_role_arn = self.iam_role.get_arn(aws_client)
            if iam_role_arn is None:
                logger.error("IamRole ARN unavailable.")
                return False
            create_response = service_client.create_crawler(
                Name=self.name,
                Role=iam_role_arn,
                **not_null_args,
            )
            logger.debug(f"GlueCrawler: {create_response}")
            logger.debug(f"GlueCrawler type: {type(create_response)}")

            if create_response is not None:
                print_info(f"GlueCrawler created: {self.name}")
                self.active_resource = create_response
                return True
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the GlueCrawler

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            get_crawler_response = service_client.get_crawler(Name=self.name)
            # logger.debug(f"GlueCrawler: {get_crawler_response}")
            # logger.debug(f"GlueCrawler type: {type(get_crawler_response)}")

            self.creation_time = get_crawler_response.get("Crawler", {}).get(
                "CreationTime", None
            )
            self.last_crawl = get_crawler_response.get("Crawler", {}).get(
                "LastCrawl", None
            )
            logger.debug(f"GlueCrawler creation_time: {self.creation_time}")
            logger.debug(f"GlueCrawler last_crawl: {self.last_crawl}")
            if self.creation_time is not None:
                logger.debug(f"GlueCrawler found: {self.name}")
                self.active_resource = get_crawler_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the GlueCrawler

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Delete the GlueCrawler
            service_client = self.get_service_client(aws_client)
            self.active_resource = None
            delete_crawler_response = service_client.delete_crawler(Name=self.name)
            # logger.debug(f"GlueCrawler: {delete_crawler_response}")
            # logger.debug(f"GlueCrawler type: {type(delete_crawler_response)}")
            print_info(f"GlueCrawler deleted: {self.name}")
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def start_crawler(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Runs the GlueCrawler

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Starting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Get the service_client
            client: AwsApiClient = aws_client or self.get_aws_client()
            service_client = self.get_service_client(client)
            # logger.debug(f"ServiceClient: {service_client}")
            # logger.debug(f"ServiceClient type: {type(service_client)}")

            try:
                start_crawler_response = service_client.start_crawler(Name=self.name)
                # logger.debug(f"start_crawler_response: {start_crawler_response}")
            except service_client.exceptions.CrawlerRunningException:
                # reference: https://github.com/boto/boto3/issues/1606
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} already running"
                )
                return True

            if start_crawler_response is not None:
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} started"
                )
                return True

        except Exception as e:
            print_error("GlueCrawler could not be started")
            print_error(e)
            logger.exception(e)
        return False



from typing import Union, Optional

from phidata.constants import AIRFLOW_EXECUTOR_ENV_VAR
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class AirflowExecutor(ExtendedEnum):
    DebugExecutor = "DebugExecutor"
    LocalExecutor = "LocalExecutor"
    SequentialExecutor = "SequentialExecutor"
    CeleryExecutor = "CeleryExecutor"
    CeleryKubernetesExecutor = "CeleryKubernetesExecutor"
    DaskExecutor = "DaskExecutor"
    KubernetesExecutor = "KubernetesExecutor"


AirflowExecutorType = Union[str, AirflowExecutor]


def get_airflow_executor() -> Optional[AirflowExecutor]:
    from os import getenv

    phidata_runtime_env_var = getenv(AIRFLOW_EXECUTOR_ENV_VAR)
    logger.debug(f"{AIRFLOW_EXECUTOR_ENV_VAR}: {phidata_runtime_env_var}")

    if (
        phidata_runtime_env_var is not None
        and phidata_runtime_env_var in AirflowExecutor.values_list()
    ):
        return AirflowExecutor.from_str(phidata_runtime_env_var)
    return AirflowExecutor.SequentialExecutor



from typing import Any, Dict, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.core.v1.config_map import ConfigMap
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.log import logger


class CreateConfigMap(BaseModel):
    cm_name: str
    app_name: str
    namespace: Optional[str] = None
    data: Optional[Dict[str, Any]] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[ConfigMap]:
        """Creates the ConfigMap resource"""

        cm_name = self.cm_name
        logger.debug(f"Init ConfigMap resource: {cm_name}")

        cm_labels = create_component_labels_dict(
            component_name=cm_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        configmap = ConfigMap(
            api_version=ApiVersion.CORE_V1,
            kind=Kind.CONFIGMAP,
            metadata=ObjectMeta(
                name=cm_name,
                namespace=self.namespace,
                labels=cm_labels,
            ),
            data=self.data,
        )

        # logger.debug(
        #     f"ConfigMap {cm_name}:\n{configmap.json(exclude_defaults=True, indent=2)}"
        # )
        return configmap



from typing import Optional, List, Union

from phidata.infra.base import InfraConfig
from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.manager import AwsManager
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.utils.log import logger


class AwsConfig(InfraConfig):
    def __init__(
        self,
        env: Optional[str] = "prd",
        version: Optional[str] = None,
        enabled: bool = True,
        # AwsResourceGroups to deploy
        resources: Optional[Union[AwsResourceGroup, List[AwsResourceGroup]]] = None,
        # Resources dir where aws manifests are stored
        resources_dir: str = "aws",
        # Aws params for this Config
        # Override the aws params from WorkspaceConfig if provided
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
    ):
        super().__init__()
        _resources: Optional[List[AwsResourceGroup]] = None
        if isinstance(resources, AwsResourceGroup):
            _resources = [resources]
        elif isinstance(resources, list):
            _resources = resources
        else:
            raise TypeError(
                "AwsConfig.resources should be AwsResourceGroup or List[AwsResourceGroup]"
            )

        try:
            self.args: AwsArgs = AwsArgs(
                env=env,
                version=version,
                enabled=enabled,
                resources=_resources,
                resources_dir=resources_dir,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def resources(self) -> Optional[List[AwsResourceGroup]]:
        return self.args.resources if self.args else None

    @property
    def resources_dir(self) -> Optional[str]:
        return self.args.resources_dir if self.args else None

    @resources_dir.setter
    def resources_dir(self, resources_dir: str) -> None:
        if self.args is not None and resources_dir is not None:
            self.args.resources_dir = resources_dir

    def is_valid(self) -> bool:
        return True

    def get_aws_manager(self) -> Optional[AwsManager]:
        return AwsManager(aws_args=self.args)



from typing import Optional

from phidata.app.phidata_app import PhidataApp, PhidataAppArgs
from phidata.utils.log import logger


class DbAppArgs(PhidataAppArgs):
    db_user: Optional[str] = None
    db_password: Optional[str] = None
    db_schema: Optional[str] = None
    db_driver: Optional[str] = None


class DbApp(PhidataApp):
    def __init__(self):
        super().__init__()
        self.args: DbAppArgs

    def get_db_user(self) -> Optional[str]:
        logger.debug(f"@get_db_user not defined for {self.__class__.__name__}")
        return self.args.db_user if self.args else None

    def get_db_password(self) -> Optional[str]:
        logger.debug(f"@get_db_password not defined for {self.__class__.__name__}")
        return self.args.db_password if self.args else None

    def get_db_schema(self) -> Optional[str]:
        logger.debug(f"@get_db_schema not defined for {self.__class__.__name__}")
        return self.args.db_schema if self.args else None

    def get_db_driver(self) -> Optional[str]:
        logger.debug(f"@get_db_driver not defined for {self.__class__.__name__}")
        return self.args.db_driver if self.args else None

    def get_db_host_local(self) -> Optional[str]:
        logger.debug(f"@get_db_host_local not defined for {self.__class__.__name__}")
        return None

    def get_db_port_local(self) -> Optional[int]:
        logger.debug(f"@get_db_port_local not defined for {self.__class__.__name__}")
        return None

    def get_db_host_docker(self) -> Optional[str]:
        logger.debug(f"@get_db_host_docker not defined for {self.__class__.__name__}")
        return None

    def get_db_port_docker(self) -> Optional[int]:
        logger.debug(f"@get_db_port_docker not defined for {self.__class__.__name__}")
        return None

    def get_db_host_k8s(self) -> Optional[str]:
        logger.debug(f"@get_db_host_k8s not defined for {self.__class__.__name__}")
        return None

    def get_db_port_k8s(self) -> Optional[int]:
        logger.debug(f"@get_db_port_k8s not defined for {self.__class__.__name__}")
        return None

    def get_db_connection_url_local(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    def get_db_connection_url_docker(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_docker()
        port = self.get_db_port_docker()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    def get_db_connection_url_k8s(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_k8s()
        port = self.get_db_port_k8s()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"



from typing import List, Optional, Dict, Any

from pydantic import BaseModel, ValidationError

from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_object import (
    CustomObject,
)
from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_resource_definition import (
    CustomResourceDefinition,
)
from phidata.infra.k8s.resource.apps.v1.deployment import Deployment
from phidata.infra.k8s.resource.core.v1.config_map import ConfigMap
from phidata.infra.k8s.resource.core.v1.namespace import Namespace
from phidata.infra.k8s.resource.core.v1.secret import Secret
from phidata.infra.k8s.resource.core.v1.service import Service
from phidata.infra.k8s.resource.core.v1.service_account import ServiceAccount
from phidata.infra.k8s.resource.core.v1.persistent_volume import PersistentVolume
from phidata.infra.k8s.resource.core.v1.persistent_volume_claim import (
    PersistentVolumeClaim,
)
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    ClusterRoleBinding,
)
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
    ClusterRole,
)
from phidata.infra.k8s.resource.storage_k8s_io.v1.storage_class import StorageClass
from phidata.utils.log import logger


class K8sResourceGroup(BaseModel):
    """
    The K8sResourceGroup is a collection of K8s Resources acting as a single unit.

    1 PhidataApp normally produces 1 K8sResourceGroup
    but for complex Apps, 1 PhidataApp may produce > 1 K8sResourceGroup
    """

    # Name of this group aka name of the app
    name: str
    enabled: bool = True

    # The weight variable controls how this is group is deployed to a cluster
    # relative to other resource groups.

    # Within each resource group, different types of resources are
    # deployed in a predefined order (eg: ns first, then sa and so on ...)
    # but we can also add an order to how to deploy different resource groups.
    # Eg: if we want to deploy a resource group with databases before other resources

    # Weights 1-10 are reserved
    # Weight 100 is default.
    # ResourceGroups with weight 100 are the default resources.
    # Choose weight 11-99 to deploy a resource group before all the default resources.
    # Choose weight 101+ to deploy a resource group after all the default resources
    weight: int = 100

    ns: Optional[List[Namespace]] = None
    sa: Optional[List[ServiceAccount]] = None
    cr: Optional[List[ClusterRole]] = None
    crb: Optional[List[ClusterRoleBinding]] = None
    secrets: Optional[List[Secret]] = None
    config_maps: Optional[List[ConfigMap]] = None
    storage_classes: Optional[List[StorageClass]] = None
    services: Optional[List[Service]] = None
    deployments: Optional[List[Deployment]] = None
    custom_objects: Optional[List[CustomObject]] = None
    crds: Optional[List[CustomResourceDefinition]] = None
    pvs: Optional[List[PersistentVolume]] = None
    pvcs: Optional[List[PersistentVolumeClaim]] = None

    def add_manifest_to_group(self, manifest: Dict[str, Any]) -> bool:
        """
        Reads a k8s manifest dict and adds to the K8sResourceGroup
        """
        kind = manifest.get("kind", None)
        if kind is None:
            logger.error(f"Cannot parse manifest: {manifest}")
            return False

        logger.debug(f"Adding {kind} to {self.name}")

        try:
            ######################################################
            ## Parse Namespace
            ######################################################
            if kind == "Namespace":
                from phidata.infra.k8s.resource.core.v1.namespace import Namespace

                _ns_resource = Namespace(**manifest)
                if _ns_resource is not None:
                    if self.ns is None:
                        self.ns = []
                    self.ns.append(_ns_resource)
                logger.debug(
                    f"Parsed: {_ns_resource.get_resource_type()}: {_ns_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse ServiceAccount
            ######################################################
            elif kind == "ServiceAccount":
                from phidata.infra.k8s.resource.core.v1.service_account import (
                    ServiceAccount,
                )

                _sa_resource = ServiceAccount(**manifest)
                if _sa_resource is not None:
                    if self.sa is None:
                        self.sa = []
                    self.sa.append(_sa_resource)
                logger.debug(
                    f"Parsed: {_sa_resource.get_resource_type()}: {_sa_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse ClusterRole
            ######################################################
            elif kind == "ClusterRole":
                from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
                    ClusterRole,
                )

                _cr_resource = ClusterRole(**manifest)
                if _cr_resource is not None:
                    if self.cr is None:
                        self.cr = []
                    self.cr.append(_cr_resource)
                logger.debug(
                    f"Parsed: {_cr_resource.get_resource_type()}: {_cr_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse ClusterRoleBinding
            ######################################################
            elif kind == "ClusterRoleBinding":
                from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
                    ClusterRoleBinding,
                )

                _crb_resource = ClusterRoleBinding(**manifest)
                if _crb_resource is not None:
                    if self.crb is None:
                        self.crb = []
                    self.crb.append(_crb_resource)
                logger.debug(
                    f"Parsed: {_crb_resource.get_resource_type()}: {_crb_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse Secrets
            ######################################################
            elif kind == "Secret":
                from phidata.infra.k8s.resource.core.v1.secret import Secret

                _secret_resource = Secret(**manifest)
                if _secret_resource is not None:
                    if self.secrets is None:
                        self.secrets = []
                    self.secrets.append(_secret_resource)
                logger.debug(
                    f"Parsed: {_secret_resource.get_resource_type()}: {_secret_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse ConfigMaps
            ######################################################
            elif kind == "ConfigMap":
                from phidata.infra.k8s.resource.core.v1.config_map import ConfigMap

                _cm_resource = ConfigMap(**manifest)
                if _cm_resource is not None:
                    if self.config_maps is None:
                        self.config_maps = []
                    self.config_maps.append(_cm_resource)
                logger.debug(
                    f"Parsed: {_cm_resource.get_resource_type()}: {_cm_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse StorageClasses
            ######################################################
            elif kind == "StorageClass":
                from phidata.infra.k8s.resource.storage_k8s_io.v1.storage_class import (
                    StorageClass,
                )

                _sc_resource = StorageClass(**manifest)
                if _sc_resource is not None:
                    if self.storage_classes is None:
                        self.storage_classes = []
                    self.storage_classes.append(_sc_resource)
                logger.debug(
                    f"Parsed: {_sc_resource.get_resource_type()}: {_sc_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse Services
            ######################################################
            elif kind == "Service":
                from phidata.infra.k8s.resource.core.v1.service import Service

                _svc_resource = Service(**manifest)
                if _svc_resource is not None:
                    if self.services is None:
                        self.services = []
                    self.services.append(_svc_resource)
                logger.debug(
                    f"Parsed: {_svc_resource.get_resource_type()}: {_svc_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse Deployments
            ######################################################
            elif kind == "Deployment":
                from phidata.infra.k8s.resource.apps.v1.deployment import Deployment

                _deploy_resource = Deployment(**manifest)
                if _deploy_resource is not None:
                    if self.deployments is None:
                        self.deployments = []
                    self.deployments.append(_deploy_resource)
                logger.debug(
                    f"Parsed: {_deploy_resource.get_resource_type()}: {_deploy_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse CustomResourceDefinitions
            ######################################################
            if kind == "CustomResourceDefinition":
                from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_resource_definition import (
                    CustomResourceDefinition,
                )

                _crd_resource = CustomResourceDefinition(**manifest)
                if _crd_resource is not None:
                    if self.crds is None:
                        self.crds = []
                    self.crds.append(_crd_resource)
                logger.debug(
                    f"Parsed: {_crd_resource.get_resource_type()}: {_crd_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse PersistentVolumes
            ######################################################
            elif kind == "PersistentVolume":
                from phidata.infra.k8s.resource.core.v1.persistent_volume import (
                    PersistentVolume,
                )

                _pv_resource = PersistentVolume(**manifest)
                if _pv_resource is not None:
                    if self.pvs is None:
                        self.pvs = []
                    self.pvs.append(_pv_resource)
                logger.debug(
                    f"Parsed: {_pv_resource.get_resource_type()}: {_pv_resource.get_resource_name()}"
                )
            ######################################################
            ## Parse PersistentVolumeClaim
            ######################################################
            elif kind == "PersistentVolumeClaim":
                from phidata.infra.k8s.resource.core.v1.persistent_volume_claim import (
                    PersistentVolumeClaim,
                )

                _pvcs_resource = PersistentVolumeClaim(**manifest)
                if _pvcs_resource is not None:
                    if self.pvcs is None:
                        self.pvcs = []
                    self.pvcs.append(_pvcs_resource)
                logger.debug(
                    f"Parsed: {_pvcs_resource.get_resource_type()}: {_pvcs_resource.get_resource_name()}"
                )
            ######################################################
            ## TODO: Parse CustomObjects
            ######################################################
            return True
        except ValidationError as validation_error:
            logger.warning(f"Could not parse manifest: {manifest}")
            logger.exception(validation_error)
            return False


class K8sBuildContext(BaseModel):
    namespace: str = "default"
    context: Optional[str] = None
    service_account_name: Optional[str] = None
    labels: Optional[Dict[str, str]] = None



from typing import Optional, Any, Dict, List

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class DbSubnetGroup(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html#RDS.Client.create_db_subnet_group

    Creates a new DB subnet group. DB subnet groups must contain at least one subnet in at least
    two AZs in the Amazon Web Services Region.
    """

    resource_type = "DbSubnetGroup"
    service_name = "rds"

    # The name for the DB subnet group. This value is stored as a lowercase string.
    # Constraints:
    #   Must contain no more than 255 letters, numbers, periods, underscores, spaces, or hyphens.
    #   Must not be default.
    #   First character must be a letter.
    #   Example: mydbsubnetgroup
    name: str
    # The description for the DB subnet group.
    description: str
    # The EC2 Subnet IDs for the DB subnet group.
    subnet_ids: Optional[List[str]] = None
    # Get Subnet IDs from a VPC CloudFormationStack
    # NOTE: only gets privatesubnets from the vpc stack
    vpc_stack: Optional[CloudFormationStack] = None
    # Tags to assign to the DB subnet group.
    tags: Optional[List[Dict[str, str]]] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the DbSubnetGroup

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}

            if self.tags:
                not_null_args["Tags"] = self.tags

            subnet_ids = self.subnet_ids
            if subnet_ids is None and self.vpc_stack is not None:
                logger.debug("Getting private subnet_ids from vpc stack")
                subnet_ids = self.vpc_stack.get_private_subnets(aws_client=aws_client)

            # Create DbSubnetGroup
            service_client = self.get_service_client(aws_client)
            create_response = service_client.create_db_subnet_group(
                DBSubnetGroupName=self.name,
                DBSubnetGroupDescription=self.description,
                SubnetIds=subnet_ids,
                **not_null_args,
            )
            logger.debug(f"create_response type: {type(create_response)}")
            logger.debug(f"create_response: {create_response}")

            self.active_resource = create_response.get("DBSubnetGroup", None)
            if self.active_resource is not None:
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} created"
                )
                logger.debug(f"DbSubnetGroup: {self.active_resource}")
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the DbSubnetGroup

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            describe_response = service_client.describe_db_subnet_groups(
                DBSubnetGroupName=self.name
            )
            logger.debug(f"describe_response type: {type(describe_response)}")
            logger.debug(f"describe_response: {describe_response}")

            db_subnet_group_list = describe_response.get("DBSubnetGroups", None)
            if db_subnet_group_list is not None and isinstance(
                db_subnet_group_list, list
            ):
                for _db_subnet_group in db_subnet_group_list:
                    _db_sg_name = _db_subnet_group.get("DBSubnetGroupName", None)
                    if _db_sg_name == self.name:
                        self.active_resource = _db_subnet_group
                        break

            if self.active_resource is None:
                logger.debug(f"No {self.get_resource_type()} found")
                return None

            logger.debug(f"DbSubnetGroup: {self.active_resource}")
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the DbSubnetGroup

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            self.active_resource = None

            delete_response = service_client.delete_db_subnet_group(
                DBSubnetGroupName=self.name
            )
            logger.debug(f"delete_response: {delete_response}")
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False



from typing import Optional, Any, Dict, List
from textwrap import dedent

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class EksFargateProfile(AwsResource):
    """
    The Fargate profile allows an administrator to declare which pods run on Fargate and specify which pods
    run on which Fargate profile. This declaration is done through the profile’s selectors.
    Each profile can have up to five selectors that contain a namespace and labels.
    A namespace is required for every selector. The label field consists of multiple optional key-value pairs.
    Pods that match the selectors are scheduled on Fargate.
    If a to-be-scheduled pod matches any of the selectors in the Fargate profile, then that pod is run on Fargate.

    fargate_role:
    When you create a Fargate profile, you must specify a pod execution role to use with the pods that are scheduled
    with the profile. This role is added to the cluster's Kubernetes Role Based Access Control (RBAC) for
    authorization so that the kubelet that is running on the Fargate infrastructure can register with your
    Amazon EKS cluster so that it can appear in your cluster as a node. The pod execution role also provides
    IAM permissions to the Fargate infrastructure to allow read access to Amazon ECR image repositories.

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks.html
    """

    resource_type = "EksFargateProfile"
    service_name = "eks"

    # Name for the fargate profile
    name: str
    # The cluster to create the EksFargateProfile in
    eks_cluster: EksCluster

    # If role is None, a default fargate_role is created using fargate_role_name
    fargate_role: Optional[IamRole] = None
    # Name for the default fargate_role when role is None, use "name-iam-role" if not provided
    fargate_role_name: Optional[str] = None

    # The Kubernetes namespace that the selector should match.
    namespace: str = "default"
    # The Kubernetes labels that the selector should match.
    # A pod must contain all of the labels that are specified in the selector for it to be considered a match.
    labels: Optional[Dict[str, str]] = None
    # Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
    # This field is autopopulated if not provided.
    client_request_token: Optional[str] = None
    # The metadata to apply to the Fargate profile to assist with categorization and organization.
    # Each tag consists of a key and an optional value. You define both.
    tags: Optional[Dict[str, str]] = None

    skip_delete = False
    # bump the wait time for Eks to 30 seconds
    waiter_delay = 30

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates a Fargate profile for your Amazon EKS cluster.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Create the Fargate IamRole if needed
            fargate_iam_role = self.get_fargate_iam_role()
            try:
                print_info(f"Creating IamRole: {fargate_iam_role.name}")
                fargate_iam_role.create(aws_client)
                fargate_iam_role_arn = fargate_iam_role.read(aws_client).arn
                print_info(f"fargate_iam_role_arn: {fargate_iam_role_arn}")
            except Exception as e:
                print_error("IamRole creation failed, please try again")
                print_error(e)
                return False

            # Get private subnets
            # Only private subnets are supported for pods that are running on Fargate.
            eks_vpc_stack: CloudFormationStack = self.eks_cluster.get_eks_vpc_stack()
            private_subnets: Optional[List[str]] = eks_vpc_stack.get_private_subnets(
                aws_client
            )

            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}
            if private_subnets is not None:
                not_null_args["subnets"] = private_subnets

            default_selector: Dict[str, Any] = {
                "namespace": self.namespace,
            }
            if self.labels is not None:
                default_selector["labels"] = self.labels
            if self.client_request_token:
                not_null_args["clientRequestToken"] = self.client_request_token
            if self.tags:
                not_null_args["tags"] = self.tags

            ## Create a Fargate profile
            # Get the service_client
            service_client = self.get_service_client(aws_client)
            # logger.debug(f"ServiceClient: {service_client}")
            # logger.debug(f"ServiceClient type: {type(service_client)}")
            try:
                print_info(f"Creating EksFargateProfile: {self.name}")
                create_profile_response = service_client.create_fargate_profile(
                    fargateProfileName=self.name,
                    clusterName=self.eks_cluster.name,
                    podExecutionRoleArn=fargate_iam_role_arn,
                    selectors=[default_selector],
                    **not_null_args,
                )
                # logger.debug(f"create_profile_response: {create_profile_response}")
                # logger.debug(
                #     f"create_profile_response type: {type(create_profile_response)}"
                # )
                ## Validate Fargate role creation
                fargate_profile_creation_time = create_profile_response.get(
                    "fargateProfile", {}
                ).get("createdAt", None)
                fargate_profile_status = create_profile_response.get(
                    "fargateProfile", {}
                ).get("status", None)
                logger.debug(f"creation_time: {fargate_profile_creation_time}")
                logger.debug(f"cluster_status: {fargate_profile_status}")
                if fargate_profile_creation_time is not None:
                    print_info(f"EksFargateProfile created: {self.name}")
                    self.active_resource = create_profile_response
                    self.active_resource_class = create_profile_response.__class__
                    return True
            except Exception as e:
                print_error(
                    "EksFargateProfile could not be created, this operation is known to be buggy."
                )
                print_error("Please deploy the workspace again.")
                print_error(e)
                return False
        except Exception as e:
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        ## Wait for EksFargateProfile to be created
        if self.wait_for_creation:
            try:
                print_info(
                    "Waiting for EksFargateProfile to be created, this can take upto 5 minutes"
                )
                waiter = self.get_service_client(aws_client).get_waiter(
                    "fargate_profile_active"
                )
                waiter.wait(
                    clusterName=self.eks_cluster.name,
                    fargateProfileName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error(
                    "Received errors while waiting for EksFargateProfile creation, this operation is known to be buggy."
                )
                print_error(e)
                return False
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EksFargateProfile

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            describe_profile_response = service_client.describe_fargate_profile(
                clusterName=self.eks_cluster.name,
                fargateProfileName=self.name,
            )
            # logger.debug(f"describe_profile_response: {describe_profile_response}")
            # logger.debug(f"describe_profile_response type: {type(describe_profile_response)}")

            fargate_profile_creation_time = describe_profile_response.get(
                "fargateProfile", {}
            ).get("createdAt", None)
            fargate_profile_status = describe_profile_response.get(
                "fargateProfile", {}
            ).get("status", None)
            logger.debug(
                f"FargateProfile creation_time: {fargate_profile_creation_time}"
            )
            logger.debug(f"FargateProfile status: {fargate_profile_status}")
            if fargate_profile_creation_time is not None:
                logger.debug(f"EksFargateProfile found: {self.name}")
                self.active_resource = describe_profile_response
                self.active_resource_class = describe_profile_response.__class__
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EksFargateProfile

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Create the Fargate IamRole
            fargate_iam_role = self.get_fargate_iam_role()
            try:
                print_info(f"Deleting IamRole: {fargate_iam_role.name}")
                fargate_iam_role.delete(aws_client)
            except Exception as e:
                print_error(
                    "IamRole deletion failed, please try again or delete manually"
                )
                print_error(e)

            # Delete the Fargate profile
            service_client = self.get_service_client(aws_client)
            self.active_resource = None
            self.active_resource_class = None
            delete_profile_response = service_client.delete_fargate_profile(
                clusterName=self.eks_cluster.name,
                fargateProfileName=self.name,
            )
            # logger.debug(f"delete_profile_response: {delete_profile_response}")
            # logger.debug(
            #     f"delete_profile_response type: {type(delete_profile_response)}"
            # )
            print_info(f"EksFargateProfile deleted: {self.name}")
            return True

        except Exception as e:
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        ## Wait for EksFargateProfile to be deleted
        if self.wait_for_deletion:
            try:
                print_info(
                    "Waiting for EksFargateProfile to be deleted, this can take upto 5 minutes"
                )
                waiter = self.get_service_client(aws_client).get_waiter(
                    "fargate_profile_deleted"
                )
                waiter.wait(
                    clusterName=self.eks_cluster.name,
                    fargateProfileName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
                return True
            except Exception as e:
                print_error(
                    "Received errors while waiting for EksFargateProfile deletion, this operation is known to be buggy."
                )
                print_error("Please try again or delete resources manually.")
                print_error(e)
        return True

    def get_fargate_iam_role(self) -> IamRole:
        """
        Create an IAM role and attach the required Amazon EKS IAM managed policy to it.
        When your cluster creates pods on Fargate infrastructure, the components running on the Fargate
        infrastructure need to make calls to AWS APIs on your behalf to do things like pull
        container images from Amazon ECR or route logs to other AWS services.
        The Amazon EKS pod execution role provides the IAM permissions to do this.
        Returns:

        """
        if self.fargate_role is not None:
            return self.fargate_role
        return IamRole(
            name=self.fargate_role_name or f"{self.name}-iam-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "eks-fargate-pods.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policy_arns=[
                "arn:aws:iam::aws:policy/AmazonEKSFargatePodExecutionRolePolicy"
            ],
        )



from collections import OrderedDict
from typing import Optional, List, Dict, Tuple

from phidata.infra.docker.args import DockerArgs
from phidata.infra.docker.api_client import DockerApiClient
from phidata.infra.docker.resource.base import DockerResource
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.docker.resource.utils import (
    filter_and_flatten_docker_resource_groups,
    get_docker_resources_from_group,
)
from phidata.infra.docker.exceptions import (
    DockerArgsException,
    DockerApiClientException,
    DockerResourceCreationFailedException,
)
from phidata.utils.common import is_empty
from phidata.utils.cli_console import (
    print_info,
    print_info,
    print_heading,
    print_subheading,
    print_error,
    print_warning,
    confirm_yes_no,
)
from phidata.utils.log import logger


class DockerWorker:
    """This class interacts with the Docker API."""

    def __init__(self, docker_args: DockerArgs) -> None:
        self.docker_args: DockerArgs = docker_args
        self.docker_client: DockerApiClient = DockerApiClient(
            base_url=self.docker_args.base_url
        )
        logger.debug(f"**-+-** DockerWorker created")

    def is_client_initialized(self) -> bool:
        return self.docker_client.is_initialized()

    def are_resources_active(self) -> bool:
        # TODO: fix this
        return False

    ######################################################
    ## Build Resources
    ######################################################

    def build_docker_resource_groups(
        self, app_filter: Optional[str] = None
    ) -> Optional[Dict[str, DockerResourceGroup]]:
        """
        Build the DockerResourceGroups for the requested apps

        Step 1: Convert each PhidataApp to DockerResourceGroups.
        Step 2: Convert any default resources from DockerConfig to a DockerResourceGroup
        Step 3: Add all DockerResourceGroups to the self.docker_resources dict
        """
        logger.debug("-*- Initializing DockerResourceGroups")

        docker_resource_groups: Optional[Dict[str, DockerResourceGroup]] = None

        # track the total number of DockerResourceGroups to build for validation
        apps_to_build = self.docker_args.apps
        resources_to_build = self.docker_args.resources

        num_apps = len(apps_to_build) if apps_to_build is not None else 0
        num_resources = len(resources_to_build) if resources_to_build is not None else 0
        num_rgs_to_build = num_apps + num_resources
        num_rgs_built = 0

        # Step 1: Convert each PhidataApp to DockerResourceGroups.
        if apps_to_build is not None and isinstance(apps_to_build, list):
            for app in apps_to_build:
                if app.args is None:
                    print_error("Args for App {} are None".format(app))
                    continue

                if not app.enabled:
                    logger.debug(f"{app.name} disabled")
                    num_rgs_built += 1
                    continue

                # skip groups not matching app_filter if provided
                if app_filter is not None:
                    if app_filter.lower() not in app.name:
                        logger.debug(f"Skipping {app.name}")
                        num_rgs_built += 1
                        continue

                logger.debug("-*- App: {}".format(app.name))

                ######################################################################
                # NOTE: VERY IMPORTANT TO GET RIGHT
                # Pass down parameters from DockerArgs -> PhidataApp
                # The DockerConfig inherits these params from the WorkspaceConfig
                # 1. Pass down the paths from the WorkspaceConfig
                # 2. Pass down docker_env
                # 3. Pass down common cloud configuration. eg: aws_region, aws_profile
                # This should match phidata.infra.prep_infra_config.prep_infra_config()
                ######################################################################

                # -*- Path parameters
                app.scripts_dir = self.docker_args.scripts_dir
                app.storage_dir = self.docker_args.storage_dir
                app.meta_dir = self.docker_args.meta_dir
                app.products_dir = self.docker_args.products_dir
                app.notebooks_dir = self.docker_args.notebooks_dir
                app.workflows_dir = self.docker_args.workflows_dir
                # The ws_root_path is the ROOT directory for the workspace
                app.workspace_root_path = self.docker_args.workspace_root_path
                app.workspace_config_dir = self.docker_args.workspace_config_dir
                app.workspace_config_file_path = (
                    self.docker_args.workspace_config_file_path
                )

                # -*- Environment parameters
                # only update the params if they are not available on the app.
                # so we can prefer the app param if provided
                if app.docker_env is None and self.docker_args.docker_env is not None:
                    app.docker_env = self.docker_args.docker_env

                # -*- AWS parameters
                # only update the params if they are not available on the app.
                # so we can prefer the app param if provided
                if app.aws_region is None and self.docker_args.aws_region is not None:
                    app.aws_region = self.docker_args.aws_region
                if app.aws_profile is None and self.docker_args.aws_profile is not None:
                    app.aws_profile = self.docker_args.aws_profile
                if (
                    app.aws_config_file is None
                    and self.docker_args.aws_config_file is not None
                ):
                    app.aws_config_file = self.docker_args.aws_config_file
                if (
                    app.aws_shared_credentials_file is None
                    and self.docker_args.aws_shared_credentials_file is not None
                ):
                    app.aws_shared_credentials_file = (
                        self.docker_args.aws_shared_credentials_file
                    )

                app_rgs: Optional[
                    Dict[str, DockerResourceGroup]
                ] = app.get_docker_resource_groups(
                    docker_build_context=DockerBuildContext(
                        network=self.docker_args.network
                    )
                )
                if app_rgs is not None:
                    if docker_resource_groups is None:
                        docker_resource_groups = OrderedDict()
                    docker_resource_groups.update(app_rgs)
                    num_rgs_built += 1

        # Step 2: Convert any default resources from DockerConfig to a DockerResourceGroup
        if self.docker_args.default_resources_available():
            logger.debug("Adding default DockerResourceGroup")
            num_rgs_to_build += 1
            default_docker_rg = DockerResourceGroup(
                # weight=10,
                images=self.docker_args.images,
                containers=self.docker_args.containers,
                volumes=self.docker_args.volumes,
            )
            if self.docker_args.resources is None:
                self.docker_args.resources = []
            self.docker_args.resources.append(default_docker_rg)

        # Step 3: Add all DockerResourceGroups to the docker_resource_groups dict
        if self.docker_args.resources is not None and isinstance(
            self.docker_args.resources, list
        ):
            for resource in self.docker_args.resources:
                if not resource.enabled:
                    logger.debug(f"{resource.name} disabled")
                    num_rgs_built += 1
                    continue

                # skip groups not matching app_filter if provided
                if app_filter is not None:
                    if app_filter.lower() not in resource.name:
                        logger.debug(f"Skipping {resource.name}")
                        num_rgs_built += 1
                        continue

                logger.debug("-*- Resource: {}".format(resource.name))

                if isinstance(resource, DockerResourceGroup):
                    if docker_resource_groups is None:
                        docker_resource_groups = OrderedDict()
                    docker_resource_groups[resource.name] = resource
                    num_rgs_built += 1

        logger.debug(
            f"# DockerResourceGroups built: {num_rgs_built}/{num_rgs_to_build}"
        )
        return docker_resource_groups

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Creating DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return True

        docker_resources_to_create: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        # Validate resources to be created
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in docker_resources_to_create:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            print_info(f"\nNetwork: {self.docker_args.network}")
            print_info(f"\nTotal {len(docker_resources_to_create)} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping deploy")
                print_info("-*-")
                exit(0)

        # track the total number of DockerResources to create for validation
        num_resources_to_create: int = len(docker_resources_to_create)
        num_resources_created: int = 0

        for resource in docker_resources_to_create:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_created = resource.create(docker_client=self.docker_client)
                if _resource_created:
                    num_resources_created += 1
                    print_info("Resource created")
                else:
                    logger.error(
                        f"Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                    )
                    if not self.docker_args.continue_on_create_failure:
                        return False
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources created: {num_resources_created}/{num_resources_to_create}"
        )
        if num_resources_to_create == num_resources_created:
            return True

        logger.error(
            f"Resources created: {num_resources_created} do not match resources required: {num_resources_to_create}"
        )
        return False

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Creating DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return

        docker_resources_to_create: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        num_resources_to_create: int = len(docker_resources_to_create)
        print_heading("--**-- Docker resources:")
        for resource in docker_resources_to_create:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        print_info(f"\nNetwork: {self.docker_args.network}")
        print_info(f"\nTotal {num_resources_to_create} resources")

    ######################################################
    ## Get Resources
    ######################################################

    def get_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[DockerResource]]:

        logger.debug("-*- Getting DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return None

        docker_resources: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        return docker_resources

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Deleting DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return True

        docker_resources_to_delete: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            sort_order="delete",
        )

        # Validate resources to be deleted
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in docker_resources_to_delete:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            print_info(f"\nNetwork: {self.docker_args.network}")
            print_info(f"\nTotal {len(docker_resources_to_delete)} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping delete")
                print_info("-*-")
                exit(0)

        # track the total number of DockerResources to delete for validation
        num_resources_to_delete: int = len(docker_resources_to_delete)
        num_resources_deleted: int = 0

        for resource in docker_resources_to_delete:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_deleted = resource.delete(docker_client=self.docker_client)
                if _resource_deleted:
                    num_resources_deleted += 1
                    print_info("Resource deleted")
                else:
                    logger.error(
                        f"Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                    )
                    if not self.docker_args.continue_on_delete_failure:
                        return False
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources deleted: {num_resources_deleted}/{num_resources_to_delete}"
        )
        if num_resources_to_delete == num_resources_deleted:
            return True

        logger.error(
            f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
        )
        return False

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Deleting DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return

        docker_resources_to_delete: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            sort_order="delete",
        )

        num_resources_to_delete: int = len(docker_resources_to_delete)
        print_heading("--**-- Docker resources:")
        for resource in docker_resources_to_delete:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        print_info(f"\nNetwork: {self.docker_args.network}")
        print_info(f"\nTotal {num_resources_to_delete} resources")

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:
        logger.debug("-*- Patching DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return True

        docker_resources_to_patch: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        # Validate resources to be patched
        if not auto_confirm:
            print_heading("--**-- Confirm resources:")
            for resource in docker_resources_to_patch:
                print_info(
                    f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
            print_info(f"\nNetwork: {self.docker_args.network}")
            print_info(f"\nTotal {len(docker_resources_to_patch)} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping patch")
                print_info("-*-")
                exit(0)

        # track the total number of DockerResources to patch for validation
        num_resources_to_patch: int = len(docker_resources_to_patch)
        num_resources_patched: int = 0

        for resource in docker_resources_to_patch:
            print_info(
                f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
            # logger.debug(resource)
            try:
                _resource_patched = resource.update(docker_client=self.docker_client)
                if _resource_patched:
                    num_resources_patched += 1
                    print_info("Resource patched")
                else:
                    logger.error(
                        f"Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                    )
                    if not self.docker_args.continue_on_patch_failure:
                        return False
            except Exception as e:
                logger.error(
                    f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                )
                logger.error("Error: {}".format(e))
                logger.error("Skipping resource creation, please fix and try again...")

        print_info(
            f"\n# Resources patched: {num_resources_patched}/{num_resources_to_patch}"
        )
        if num_resources_to_patch == num_resources_patched:
            return True

        logger.error(
            f"Resources patched: {num_resources_patched} do not match resources required: {num_resources_to_patch}"
        )
        return False

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:
        logger.debug("-*- Patching DockerResources")

        docker_resource_groups: Optional[
            Dict[str, DockerResourceGroup]
        ] = self.build_docker_resource_groups(app_filter=app_filter)

        if docker_resource_groups is None:
            print_info("No resources available")
            return

        docker_resources_to_patch: List[
            DockerResource
        ] = filter_and_flatten_docker_resource_groups(
            docker_resource_groups=docker_resource_groups,
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
        )

        num_resources_to_patch: int = len(docker_resources_to_patch)
        print_heading("--**-- Docker resources:")
        for resource in docker_resources_to_patch:
            print_info(
                f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}"
            )
        print_info(f"\nNetwork: {self.docker_args.network}")
        print_info(f"\nTotal {num_resources_to_patch} resources")

    ######################################################
    ## End
    ######################################################



from collections import OrderedDict
from typing import Dict, List, Type, Union

from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_object import (
    CustomObject,
)
from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_resource_definition import (
    CustomResourceDefinition,
)
from phidata.infra.k8s.resource.apps.v1.deployment import Deployment
from phidata.infra.k8s.resource.core.v1.config_map import ConfigMap
from phidata.infra.k8s.resource.core.v1.container import Container
from phidata.infra.k8s.resource.core.v1.namespace import Namespace
from phidata.infra.k8s.resource.core.v1.persistent_volume import (
    PersistentVolume,
)
from phidata.infra.k8s.resource.core.v1.persistent_volume_claim import (
    PersistentVolumeClaim,
)
from phidata.infra.k8s.resource.core.v1.pod import Pod
from phidata.infra.k8s.resource.core.v1.secret import Secret
from phidata.infra.k8s.resource.core.v1.service import Service
from phidata.infra.k8s.resource.core.v1.service_account import ServiceAccount
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    ClusterRoleBinding,
)
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
    ClusterRole,
)
from phidata.infra.k8s.resource.storage_k8s_io.v1.storage_class import StorageClass

# Use this as a type for an object which can hold any K8s Resource
# See get_k8s_resources_from_group() function for an example
K8sResourceType = Union[
    Namespace,
    Secret,
    ConfigMap,
    StorageClass,
    PersistentVolume,
    PersistentVolumeClaim,
    ServiceAccount,
    ClusterRole,
    ClusterRoleBinding,
    # Role,
    # RoleBinding,
    Service,
    Pod,
    Deployment,
    # Ingress,
    CustomResourceDefinition,
    CustomObject,
    Container,
]

# Use this as an ordered list to iterate over all K8s Resource Classes
# This list is the order in which resources should be installed as well.
# Copied from https://github.com/helm/helm/blob/release-2.10/pkg/tiller/kind_sorter.go#L29
K8sResourceTypeList: List[Type[Union[K8sResource, K8sObject]]] = [
    Namespace,
    ServiceAccount,
    StorageClass,
    Secret,
    ConfigMap,
    PersistentVolume,
    PersistentVolumeClaim,
    ClusterRoleBinding,
    ClusterRole,
    # Role,
    # RoleBinding,
    Pod,
    Deployment,
    Container,
    Service,
    # Ingress,
    CustomResourceDefinition,
    CustomObject,
]

# Map K8s resource alias' to their type
_k8s_resource_type_names: Dict[str, Type[Union[K8sResource, K8sObject]]] = {
    k8s_type.__name__.lower(): k8s_type for k8s_type in K8sResourceTypeList
}
_k8s_resource_type_aliases: Dict[str, Type[Union[K8sResource, K8sObject]]] = {
    "crd": CustomResourceDefinition,
    "ns": Namespace,
    "cm": ConfigMap,
    "sc": StorageClass,
    "pvc": PersistentVolumeClaim,
    "sa": ServiceAccount,
    "cr": ClusterRole,
    "crb": ClusterRoleBinding,
    "svc": Service,
    "deploy": Deployment,
}

K8sResourceAliasToTypeMap: Dict[str, Type[Union[K8sResource, K8sObject]]] = dict(
    **_k8s_resource_type_names, **_k8s_resource_type_aliases
)

# Maps each K8sResource to an install weight
# lower weight K8sResource(s) get installed first
# i.e. Namespace is installed first, then Secret... and so on
K8sResourceInstallOrder: Dict[str, int] = OrderedDict(
    {
        resource_type.__name__: idx
        for idx, resource_type in enumerate(K8sResourceTypeList, start=1)
    }
)



from pathlib import Path
from typing import Optional, Any, Dict

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.k8s.resource.kubeconfig import (
    Kubeconfig,
    KubeconfigCluster,
    KubeconfigClusterConfig,
    KubeconfigContext,
    KubeconfigContextSpec,
    KubeconfigUser,
)
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class EksKubeconfig(AwsResource):

    resource_type = "Kubeconfig"
    service_name = "na"

    # Optional: kubeconfig name, used for filtering during phi ws up/down
    name: Optional[str] = None
    # Required: EksCluster to generate the kubeconfig for
    eks_cluster: EksCluster
    # Required: Path to kubeconfig file
    kubeconfig_path: Path = Path.home().joinpath(".kube").joinpath("config").resolve()

    # Optional: cluster_name to use in kubeconfig, defaults to eks_cluster.name
    kubeconfig_cluster_name: Optional[str] = None
    # Optional: cluster_user to use in kubeconfig, defaults to eks_cluster.name
    kubeconfig_cluster_user: Optional[str] = None
    # Optional: cluster_context to use in kubeconfig, defaults to eks_cluster.name
    kubeconfig_cluster_context: Optional[str] = None

    # Optional: role to assume when signing the token
    kubeconfig_role: Optional[IamRole] = None
    # Optional: role arn to assume when signing the token
    kubeconfig_role_arn: Optional[str] = None

    # Dont delete this EksKubeconfig from the kubeconfig file
    skip_delete = True
    # Mark use_cache as False so the kubeconfig is re-created
    # every time phi ws up/down is run
    use_cache: bool = False

    def get_resource_name(self) -> Optional[str]:
        return self.name if self.name else f"{self.eks_cluster.name}-kubeconfig"

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EksKubeconfig

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            return self.write_kubeconfig(aws_client=aws_client)
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Reads the EksKubeconfig

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            kubeconfig_path = self.get_kubeconfig_path()
            if kubeconfig_path is not None:
                return Kubeconfig.read_from_file(kubeconfig_path)
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Updates the EksKubeconfig

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            return self.write_kubeconfig(aws_client=aws_client)
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be updated.")
            print_error(e)
        return False

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EksKubeconfig

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            return self.clean_kubeconfig(aws_client=aws_client)
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error(e)
        return False

    def get_kubeconfig_path(self) -> Optional[Path]:
        return self.kubeconfig_path or self.eks_cluster.kubeconfig_path

    def get_kubeconfig_cluster_name(self) -> str:
        return (
            self.kubeconfig_cluster_name
            or self.eks_cluster.get_kubeconfig_cluster_name()
        )

    def get_kubeconfig_user_name(self) -> str:
        return (
            self.kubeconfig_cluster_user or self.eks_cluster.get_kubeconfig_user_name()
        )

    def get_kubeconfig_context_name(self) -> str:
        return (
            self.kubeconfig_cluster_context
            or self.eks_cluster.get_kubeconfig_context_name()
        )

    def get_kubeconfig_role(self) -> Optional[IamRole]:
        return self.kubeconfig_role or self.eks_cluster.kubeconfig_role

    def get_kubeconfig_role_arn(self) -> Optional[str]:
        return self.kubeconfig_role_arn or self.eks_cluster.kubeconfig_role_arn

    def write_kubeconfig(self, aws_client: AwsApiClient) -> bool:

        # Step 1: Get the EksCluster to generate the kubeconfig for
        eks_cluster = self.eks_cluster._read(aws_client=aws_client)  # type: ignore
        if eks_cluster is None:
            logger.warning(f"EKSCluster not available: {self.eks_cluster.name}")
            return False

        # Step 2: Get EksCluster cert, endpoint & arn
        try:
            cluster_cert = (
                eks_cluster.get("cluster", {})
                .get("certificateAuthority", {})
                .get("data", None)
            )
            logger.debug(f"cluster_cert: {cluster_cert}")

            cluster_endpoint = eks_cluster.get("cluster", {}).get("endpoint", None)
            logger.debug(f"cluster_endpoint: {cluster_endpoint}")

            cluster_arn = eks_cluster.get("cluster", {}).get("arn", None)
            logger.debug(f"cluster_arn: {cluster_arn}")
        except Exception as e:
            print_error(f"Cannot read EKSCluster")
            print_error(e)
            return False

        # Step 3: Build Kubeconfig components
        # 3.1 Build KubeconfigCluster config
        cluster_name = self.get_kubeconfig_cluster_name()
        new_cluster = KubeconfigCluster(
            name=cluster_name,
            cluster=KubeconfigClusterConfig(
                server=str(cluster_endpoint),
                certificate_authority_data=str(cluster_cert),
            ),
        )

        # 3.2 Build KubeconfigUser config
        new_user_exec_args = ["eks", "get-token", "--cluster-name", cluster_name]
        if aws_client.aws_region is not None:
            new_user_exec_args.extend(["--region", aws_client.aws_region])
        # Assume the role if the role_arn is provided
        role = self.get_kubeconfig_role()
        role_arn = self.get_kubeconfig_role_arn()
        if role_arn is not None:
            new_user_exec_args.extend(["--role-arn", role_arn])
        # Otherwise if role is provided, use that to get the role arn
        elif role is not None:
            _arn = role.get_arn(aws_client=aws_client)
            if _arn is not None:
                new_user_exec_args.extend(["--role-arn", _arn])

        new_user_exec: Dict[str, Any] = {
            "apiVersion": ApiVersion.CLIENT_AUTHENTICATION_V1BETA1.value,
            "command": "aws",
            "args": new_user_exec_args,
        }
        if aws_client.aws_profile is not None:
            new_user_exec["env"] = [
                {"name": "AWS_PROFILE", "value": aws_client.aws_profile}
            ]

        new_user = KubeconfigUser(
            name=self.get_kubeconfig_user_name(),
            user={"exec": new_user_exec},
        )

        # 3.3 Build KubeconfigContext config
        new_context = KubeconfigContext(
            name=self.get_kubeconfig_context_name(),
            context=KubeconfigContextSpec(
                cluster=new_cluster.name,
                user=new_user.name,
            ),
        )
        current_context = new_context.name
        cluster_config: KubeconfigCluster

        # Step 4: Get existing Kubeconfig
        kubeconfig_path = self.get_kubeconfig_path()
        if kubeconfig_path is None:
            logger.error(f"kubeconfig_path is None")
            return False

        kubeconfig: Optional[Any] = Kubeconfig.read_from_file(kubeconfig_path)

        # Step 5: Parse through the existing config to determine if
        # an update is required. By the end of this logic
        # if write_kubeconfig = False then no changes to kubeconfig are needed
        # if write_kubeconfig = True then we should write the kubeconfig file
        write_kubeconfig = False

        # Kubeconfig exists and is valid
        if kubeconfig is not None and isinstance(kubeconfig, Kubeconfig):
            # Update Kubeconfig.clusters:
            # If a cluster with the same name exists in Kubeconfig.clusters
            #   - check if server and cert values match, if not, remove the existing cluster
            #   and add the new cluster config. Mark cluster_config_exists = True
            # If a cluster with the same name does not exist in Kubeconfig.clusters
            #   - add the new cluster config
            cluster_config_exists = False
            for idx, _cluster in enumerate(kubeconfig.clusters, start=0):
                if _cluster.name == new_cluster.name:
                    cluster_config_exists = True
                    if (
                        _cluster.cluster.server != new_cluster.cluster.server
                        or _cluster.cluster.certificate_authority_data
                        != new_cluster.cluster.certificate_authority_data
                    ):
                        logger.debug(
                            "Kubeconfig.cluster mismatch, updating cluster config"
                        )
                        removed_cluster_config = kubeconfig.clusters.pop(idx)
                        # logger.debug(
                        #     f"removed_cluster_config: {removed_cluster_config}"
                        # )
                        kubeconfig.clusters.append(new_cluster)
                        write_kubeconfig = True
            if not cluster_config_exists:
                logger.debug("Adding Kubeconfig.cluster")
                kubeconfig.clusters.append(new_cluster)
                write_kubeconfig = True

            # Update Kubeconfig.users:
            # If a user with the same name exists in Kubeconfig.users -
            #   check if user spec matches, if not, remove the existing user
            #   and add the new user config. Mark user_config_exists = True
            # If a user with the same name does not exist in Kubeconfig.users -
            #   add the new user config
            user_config_exists = False
            for idx, _user in enumerate(kubeconfig.users, start=0):
                if _user.name == new_user.name:
                    user_config_exists = True
                    if _user.user != new_user.user:
                        logger.debug("Kubeconfig.user mismatch, updating user config")
                        removed_user_config = kubeconfig.users.pop(idx)
                        # logger.debug(f"removed_user_config: {removed_user_config}")
                        kubeconfig.users.append(new_user)
                        write_kubeconfig = True
            if not user_config_exists:
                logger.debug("Adding Kubeconfig.user")
                kubeconfig.users.append(new_user)
                write_kubeconfig = True

            # Update Kubeconfig.contexts:
            # If a context with the same name exists in Kubeconfig.contexts -
            #   check if context spec matches, if not, remove the existing context
            #   and add the new context. Mark context_config_exists = True
            # If a context with the same name does not exist in Kubeconfig.contexts -
            #   add the new context config
            context_config_exists = False
            for idx, _context in enumerate(kubeconfig.contexts, start=0):
                if _context.name == new_context.name:
                    context_config_exists = True
                    if _context.context != new_context.context:
                        logger.debug(
                            "Kubeconfig.context mismatch, updating context config"
                        )
                        removed_context_config = kubeconfig.contexts.pop(idx)
                        # logger.debug(
                        #     f"removed_context_config: {removed_context_config}"
                        # )
                        kubeconfig.contexts.append(new_context)
                        write_kubeconfig = True
            if not context_config_exists:
                logger.debug("Adding Kubeconfig.context")
                kubeconfig.contexts.append(new_context)
                write_kubeconfig = True

            if (
                kubeconfig.current_context is None
                or kubeconfig.current_context != current_context
            ):
                logger.debug("Updating Kubeconfig.current_context")
                kubeconfig.current_context = current_context
                write_kubeconfig = True
        else:
            # Kubeconfig does not exist or is not valid
            # Create a new Kubeconfig
            logger.info(f"Creating new Kubeconfig")
            kubeconfig = Kubeconfig(
                clusters=[new_cluster],
                users=[new_user],
                contexts=[new_context],
                current_context=current_context,
            )
            write_kubeconfig = True

        # if kubeconfig:
        #     logger.debug("Kubeconfig:\n{}".format(kubeconfig.json(exclude_none=True, by_alias=True, indent=4)))

        # Step 5: Write Kubeconfig if an update is made
        if write_kubeconfig:
            return kubeconfig.write_to_file(kubeconfig_path)
        else:
            logger.info("Kubeconfig up-to-date")
        return True

    def clean_kubeconfig(self, aws_client: AwsApiClient) -> bool:
        logger.debug(f"TO_DO: Cleaning kubeconfig at {str(self.kubeconfig_path)}")
        return True

    ###########################################################################
    # def write_kubeconfig(self, aws_client: AwsApiClient) -> bool:
    #
    #     # Step 1: Get the EksCluster
    #     eks_cluster = self.eks_cluster._read(aws_client=aws_client)
    #     if eks_cluster is None:
    #         logger.warning("No EKSCluster available for kubeconfig update")
    #         return False
    #
    #     # Step 2: Get EksCluster cert, endpoint & arn
    #     try:
    #         cluster_cert = (
    #             eks_cluster.get("cluster", {})
    #             .get("certificateAuthority", {})
    #             .get("data", None)
    #         )
    #         logger.debug(f"cluster_cert: {cluster_cert}")
    #         cluster_endpoint = eks_cluster.get("cluster", {}).get("endpoint", None)
    #         logger.debug(f"cluster_endpoint: {cluster_endpoint}")
    #         cluster_arn = eks_cluster.get("cluster", {}).get("arn", None)
    #         logger.debug(f"cluster_arn: {cluster_arn}")
    #     except Exception as e:
    #         print_error(f"Cannot read EKSCluster")
    #         print_error(e)
    #         return False
    #
    #     from phidata.infra.k8s.enums.api_version import ApiVersion
    #     from phidata.infra.k8s.enums.kind import Kind
    #     from phidata.infra.k8s.resource.kubeconfig import (
    #         Kubeconfig,
    #         KubeconfigCluster,
    #         KubeconfigClusterConfig,
    #         KubeconfigContext,
    #         KubeconfigContextSpec,
    #         KubeconfigUser,
    #     )
    #
    #     # Build cluster config
    #     new_cluster = KubeconfigCluster(
    #         name=self.get_kubeconfig_cluster_name(),
    #         cluster=KubeconfigClusterConfig(
    #             server=str(cluster_endpoint),
    #             certificate_authority_data=str(cluster_cert),
    #         ),
    #     )
    #
    #     # Build user config
    #     new_user_exec_args = ["eks", "get-token", "--cluster-name", self.name]
    #     if aws_client.aws_region is not None:
    #         new_user_exec_args.extend(["--region", aws_client.aws_region])
    #     # Assume the role using the role_arn if provided
    #     if self.role_arn is not None:
    #         new_user_exec_args.extend(["--role-arn", self.role_arn])
    #     # Otherwise if role is provided, use that to get the role arn
    #     elif self.role is not None:
    #         _arn = self.role.get_arn(aws_client=aws_client)
    #         if _arn is not None:
    #             new_user_exec_args.extend(["--role-arn", _arn])
    #
    #     new_user_exec: Dict[str, Any] = {
    #         "apiVersion": ApiVersion.CLIENT_AUTHENTICATION_V1BETA1.value,
    #         "command": "aws",
    #         "args": new_user_exec_args,
    #     }
    #     if aws_client.aws_profile is not None:
    #         new_user_exec["env"] = [
    #             {"name": "AWS_PROFILE", "value": aws_client.aws_profile}
    #         ]
    #
    #     new_user = KubeconfigUser(
    #         name=self.get_kubeconfig_user_name(),
    #         user={"exec": new_user_exec},
    #     )
    #
    #     # Build context config
    #     new_context = KubeconfigContext(
    #         name=self.get_kubeconfig_context_name(),
    #         context=KubeconfigContextSpec(
    #             cluster=new_cluster.name,
    #             user=new_user.name,
    #         ),
    #     )
    #     current_context = new_context.name
    #     cluster_config: KubeconfigCluster
    #
    #     # Step 3: Read existing Kubeconfig
    #     kubeconfig = self._read(aws_client=aws_client)
    #
    #     # Step 4: Parse through the existing config to determine if
    #     # an update is required. By the end of this logic if write_kubeconfig = False
    #     # then no changes to kubeconfig are needed
    #     write_kubeconfig = False
    #
    #     if kubeconfig is None or not isinstance(kubeconfig, Kubeconfig):
    #         logger.info(f"Creating new Kubeconfig")
    #         kubeconfig = Kubeconfig(
    #             api_version=ApiVersion.CORE_V1,
    #             kind=Kind.CONFIG,
    #             clusters=[new_cluster],
    #             users=[new_user],
    #             contexts=[new_context],
    #             current_context=current_context,
    #         )
    #         write_kubeconfig = True
    #     else:
    #         # Update Kubeconfig.clusters:
    #         # If a cluster with the same name exists in Kubeconfig.clusters
    #         #   - check if server and cert values match, if not, remove the existing cluster
    #         #   and add the new cluster config. Mark cluster_config_exists = True
    #         # If a cluster with the same name does not exist in Kubeconfig.clusters
    #         #   - add the new cluster config
    #         cluster_config_exists = False
    #         for idx, _cluster in enumerate(kubeconfig.clusters, start=0):
    #             if _cluster.name == new_cluster.name:
    #                 cluster_config_exists = True
    #                 if (
    #                     _cluster.cluster.server != new_cluster.cluster.server
    #                     or _cluster.cluster.certificate_authority_data
    #                     != new_cluster.cluster.certificate_authority_data
    #                 ):
    #                     logger.debug(
    #                         "Kubeconfig.cluster mismatch, updating cluster config"
    #                     )
    #                     removed_cluster_config = kubeconfig.clusters.pop(idx)
    #                     # logger.debug(
    #                     #     f"removed_cluster_config: {removed_cluster_config}"
    #                     # )
    #                     kubeconfig.clusters.append(new_cluster)
    #                     write_kubeconfig = True
    #         if not cluster_config_exists:
    #             logger.debug("Adding Kubeconfig.cluster")
    #             kubeconfig.clusters.append(new_cluster)
    #             write_kubeconfig = True
    #
    #         # Update Kubeconfig.users:
    #         # If a user with the same name exists in Kubeconfig.users -
    #         #   check if user spec matches, if not, remove the existing user
    #         #   and add the new user config. Mark user_config_exists = True
    #         # If a user with the same name does not exist in Kubeconfig.users -
    #         #   add the new user config
    #         user_config_exists = False
    #         for idx, _user in enumerate(kubeconfig.users, start=0):
    #             if _user.name == new_user.name:
    #                 user_config_exists = True
    #                 if _user.user != new_user.user:
    #                     logger.debug("Kubeconfig.user mismatch, updating user config")
    #                     removed_user_config = kubeconfig.users.pop(idx)
    #                     # logger.debug(f"removed_user_config: {removed_user_config}")
    #                     kubeconfig.users.append(new_user)
    #                     write_kubeconfig = True
    #         if not user_config_exists:
    #             logger.debug("Adding Kubeconfig.user")
    #             kubeconfig.users.append(new_user)
    #             write_kubeconfig = True
    #
    #         # Update Kubeconfig.contexts:
    #         # If a context with the same name exists in Kubeconfig.contexts -
    #         #   check if context spec matches, if not, remove the existing context
    #         #   and add the new context. Mark context_config_exists = True
    #         # If a context with the same name does not exist in Kubeconfig.contexts -
    #         #   add the new context config
    #         context_config_exists = False
    #         for idx, _context in enumerate(kubeconfig.contexts, start=0):
    #             if _context.name == new_context.name:
    #                 context_config_exists = True
    #                 if _context.context != new_context.context:
    #                     logger.debug(
    #                         "Kubeconfig.context mismatch, updating context config"
    #                     )
    #                     removed_context_config = kubeconfig.contexts.pop(idx)
    #                     # logger.debug(
    #                     #     f"removed_context_config: {removed_context_config}"
    #                     # )
    #                     kubeconfig.contexts.append(new_context)
    #                     write_kubeconfig = True
    #         if not context_config_exists:
    #             logger.debug("Adding Kubeconfig.context")
    #             kubeconfig.contexts.append(new_context)
    #             write_kubeconfig = True
    #
    #         if (
    #             kubeconfig.current_context is None
    #             or kubeconfig.current_context != current_context
    #         ):
    #             logger.debug("Updating Kubeconfig.current_context")
    #             kubeconfig.current_context = current_context
    #             write_kubeconfig = True
    #
    #     # if kubeconfig:
    #     #     logger.debug("Kubeconfig:\n{}".format(kubeconfig.json(exclude_none=True, by_alias=True, indent=4)))
    #
    #     # Step 5: Write Kubeconfig if an update is made
    #     if write_kubeconfig:
    #         return self._write(kubeconfig)
    #     else:
    #         logger.info("Kubeconfig up-to-date")
    #     return True
    #
    # def clean_kubeconfig(self, aws_client: AwsApiClient) -> bool:
    #     logger.debug(f"TO_DO: Cleaning kubeconfig at {str(self.kubeconfig_path)}")
    #     return True



from datetime import datetime
from pathlib import Path
from typing import Optional, Dict, Any

from pydantic import BaseModel


class RunContext(BaseModel):
    # Run specific variables
    run_date: datetime
    dry_run: bool = False
    detach: bool = False
    run_status: bool = False
    run_env: Optional[str] = None
    run_env_vars: Optional[Dict[str, str]] = None
    run_params: Optional[Dict[str, str]] = None


class PathContext(BaseModel):
    # Env specific path variables - their values are different on
    # local, docker or cloud environments.
    # These are updated by `phi wf run` for local runs
    # And are provided as Environment variables on containers

    scripts_dir: Optional[Path] = None
    storage_dir: Optional[Path] = None
    meta_dir: Optional[Path] = None
    products_dir: Optional[Path] = None
    notebooks_dir: Optional[Path] = None
    workspace_config_dir: Optional[Path] = None
    workflow_file: Optional[Path] = None


class ContainerPathContext(BaseModel):
    workspace_name: Optional[str] = None
    workspace_root: Optional[str] = None
    workspace_parent: Optional[str] = None
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    meta_dir: Optional[str] = None
    products_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    workflows_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None
    requirements_file: Optional[str] = None


class AirflowContext(BaseModel):
    ds: Optional[Any] = None
    ds_nodash: Optional[Any] = None
    logical_date: Optional[Any] = None
    data_interval_start: Optional[Any] = None
    data_interval_end: Optional[Any] = None

    ts: Optional[Any] = None
    ts_nodash: Optional[Any] = None

    test_mode: Optional[Any] = None
    conn: Optional[Any] = None

    conf: Optional[Any] = None
    dag: Optional[Any] = None
    dag_run: Optional[Any] = None
    task: Optional[Any] = None
    task_instance: Optional[Any] = None


class DockerContext(BaseModel):
    pass


class K8sContext(BaseModel):
    pass
